# -*- coding: utf-8 -*-
from decorated.base.function import Function
from decorated.util import reporters
from decorated.util.gcutil import DisableGc
import doctest
import sys
import time

if sys.platform == "win32":
    # On Windows, the best timer is time.clock()
    timer = time.clock
else:
    # On most other platforms the best timer is time.time()
    timer = time.time

class TimeIt(Function):
    def _call(self, *args, **kw):
        timings = [None] * self._repeats
        with DisableGc():
            for i in range(self._repeats):
                begin = timer()
                for _ in range(self._iterations):
                    result = super(TimeIt, self)._call(*args, **kw)
                seconds = timer() - begin
                timings[i] = seconds
        timing = Result(self._func, args, kw, self._iterations,
                self._repeats, timings)
        self._reporter(timing)
        return result
    
    def _init(self, iterations=1, repeats=1, reporter=reporters.PRINT_REPORTER):
        super(TimeIt, self)._init()
        self._iterations = iterations
        self._repeats = repeats
        self._reporter = reporter
        
class Result(object):
    def __init__(self, func, args, kwargs, iterations, repeats, timings):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.iterations = iterations
        self.repeats = repeats
        self.timings = timings
        
    def __str__(self):
        '''
        basic
        >>> def foo(): pass
        >>> foo.__module__ = 'aaa.bbb'
        >>> str(Result(foo, [], {}, 10, 3, [0.05, 0.1, 0.15]))
        'Timing of aaa.bbb.foo(): [0.05, 0.1, 0.15] (iterations=10)'
        
        with args
        >>> def foo(a): pass
        >>> foo.__module__ = 'aaa.bbb'
        >>> str(Result(foo, [1], {}, 10, 3, [0.05, 0.1, 0.15]))
        'Timing of aaa.bbb.foo(1): [0.05, 0.1, 0.15] (iterations=10)'
        
        with kwargs
        >>> def foo(a, b=0): pass
        >>> foo.__module__ = 'aaa.bbb'
        >>> str(Result(foo, [1], {'b': 2}, 10, 3, [0.05, 0.1, 0.15]))
        'Timing of aaa.bbb.foo(1, b=2): [0.05, 0.1, 0.15] (iterations=10)'
        '''
        kwargs = ['%s=%s' % (k, v) for k, v in self.kwargs.items()]
        args = [str(arg) for arg in self.args] + kwargs
        args = ', '.join(args)
        return 'Timing of %s.%s(%s): %s (iterations=%d)' \
                % (self.func.__module__, self.func.__name__, args, self.timings, self.iterations)
    
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base.function import ContextFunction
import signal
import time

class Timeout(ContextFunction):
    def _after(self, ret, *args, **kw):
        self._restore()
        
    def _before(self, *args, **kw):
        if self._seconds is None or self._seconds == 0:
            return
        
        self._old_handler = signal.getsignal(signal.SIGALRM)
        def _timeout(*args):
            raise TimeoutError()
        signal.signal(signal.SIGALRM, _timeout)
        old_alarm = signal.alarm(self._seconds)
        if old_alarm != 0:
            self._old_alarm_time = time.time() + old_alarm
    
    def _error(self, error, *args, **kw):
        self._restore()
    
    def _init(self, seconds):
        self._seconds = seconds
        self._old_handler = None
        self._old_alarm_time = None
        
    def _restore(self):
        if self._seconds is None or self._seconds == 0:
            return
        
        signal.alarm(0)
        if self._old_handler is not None:
            signal.signal(signal.SIGALRM, self._old_handler)
        if self._old_alarm_time is not None:
            remain_seconds = int(self._old_alarm_time - time.time())
            if remain_seconds < 1:
                # old alarm is overdue
                # the best we can do is rescheduling it at 1 second later
                remain_seconds = 1
            signal.alarm(remain_seconds)
        
class TimeoutError(Exception):
    pass



# -*- coding: utf-8 -*-
from cProfile import Profile as RawProfile
from decorated.base.function import Function
from decorated.util import reporters
from decorated.util.gcutil import DisableGc
from pstats import Stats
from six import StringIO
import doctest

class Profile(Function):
    def _call(self, *args, **kw):
        profile = RawProfile()
        def _run():
            with DisableGc():
                for _ in range(self._iterations):
                    _run.result = super(Profile, self)._call(*args, **kw)
        profile.runctx('_run()', {}, {'_run': _run})
        profile.create_stats()
        stats = Stats(profile)
        stats.sort_stats('cumulative')
        stats.fcn_list = stats.fcn_list[:self._max_lines]
        self._reporter(stats)
        return _run.result
    
    def _init(self, iterations=1, reporter=reporters.PRINT_REPORTER, max_lines=50):
        super(Profile, self)._init()
        self._iterations = iterations
        self._reporter = reporter
        self._max_lines = max_lines
        
class Stats(Stats):
    def __str__(self):
        '''
        >>> result = str(Stats(RawProfile().run('')))
        >>> 'function calls' in result
        True
        '''
        self.stream = StringIO()
        self.print_stats()
        return self.stream.getvalue()
    
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base.function import Function
import logging
import time

log = logging.getLogger(__name__)

class Retries(Function):
    def _init(self, times, delay=0, error_types=(Exception,)):
        if times < 0:
            raise Exception('Times cannot be negative.')
        self._times = times
        self._delay = delay
        self._error_types = error_types
        
    def _call(self, *args, **kw):
        times = 0
        while True:
            try:
                return super(Retries, self)._call(*args, **kw)
            except Exception as e:
                times += 1
                if times <= self._times and isinstance(e, self._error_types):
                    log.warn('Execution failed. Will retry in %f seconds.' % self._delay, exc_info=True)
                    time.sleep(self._delay)
                else:
                    log.warn('Execution failed permanently.')
                    raise
        


# -*- coding: utf-8 -*-
from decorated.base.function import WrapperFunction
from decorated.decorators.conditional import Conditional
from decorated.decorators.once import Once
from decorated.testing import DecoratedFixture
from fixtures2 import TestCase

class TestWrapperFunction(WrapperFunction):
    called = False
    
    def _before(self, *args, **kw):
        TestWrapperFunction.called = True
        
class DisableTest(TestCase):
    def setUp(self):
        super(DisableTest, self).setUp()
        self.decorated = self.useFixture(DecoratedFixture())
        
    def test_with_args(self):
        # set up
        @Conditional('a != 0')
        def foo(a):
            return a
        self.decorated.disable(Conditional)
        
        # test
        self.assertEqual(0, foo(0))
        self.assertEqual(1, foo(1))
        
    def test_no_arg(self):
        # set up
        DisableTest.args = []
        @Once
        def foo(a):
            DisableTest.args.append(a)
            return a
        self.decorated.disable(Once)
        
        # test
        self.assertEqual(1, foo(1))
        self.assertEqual(1, foo(1))
        self.assertEqual([1, 1], self.args)
        
    def test_method(self):
        # set up
        class Foo(object):
            @Conditional('a != 0')
            def bar(self, a):
                return a
        self.decorated.disable(Conditional)
        
        # test
        self.assertEqual(0, Foo().bar(0))
        self.assertEqual(1, Foo().bar(1))
        
    def test_wrapper_function(self):
        # set up
        @TestWrapperFunction
        def foo(a):
            return a
        self.decorated.disable(TestWrapperFunction)

        # test
        foo(0)
        self.assertFalse(TestWrapperFunction.called)
        
class EnableTest(TestCase):
    def test_with_args(self):
        # set up
        self.decorated = self.useFixture(DecoratedFixture())
        @Conditional('a != 0')
        def foo(a):
            return a
        self.decorated.disable(Conditional)
        self.decorated.enable(Conditional)
        
        # test
        self.assertIsNone(foo(0))
        self.assertEqual(1, foo(1))
        


# -*- coding: utf-8 -*-
import threading

class ThreadLocal(object):
    def __init__(self, default=None):
        super(ThreadLocal, self).__init__()
        self._data = threading.local()
        self._default = default
    
    def get(self):
        try:
            return self._data.value
        except AttributeError:
            return self._default
    
    def set(self, value):
        self._data.value = value
        
    def reset(self):
        del self._data.value
    


# -*- coding: utf-8 -*-
import doctest

class TemplateError(Exception):
    pass

class Template(object):
    def __init__(self, expression, parts):
        self._expression = expression
        self._parts = parts
        
    def eval(self, values):
        parts = []
        for p in self._parts:
            try:
                part = p.eval(values)
                part = str(part)
                parts.append(part)
            except Exception:
                raise TemplateError('Failed to evaluate expression "%s" in template "%s".' % (p._expression, self._expression))
        return ''.join(parts)

class Part(object):
    def __init__(self, expression):
        self._expression = expression
        
class StringPart(Part):
    '''
    >>> StringPart('abc').eval({'a': 1, 'b': 2})
    'abc'
    '''
    def eval(self, variables):
        return self._expression
    
class VariablePart(Part):
    '''
    >>> from decorated.base.dict import Dict
    >>> VariablePart('"abc"').eval({'a': 1, 'b': 2})
    'abc'
    >>> VariablePart('a').eval({'a': 1, 'b': 2})
    1
    >>> VariablePart('user.id').eval({'user': Dict(id=1)})
    1
    >>> VariablePart('user["id"]').eval({'user': {'id': 1}})
    1
    '''
    def eval(self, variables):
        return eval(self._expression, variables)

def compile(template, names):
    parts = []
    expression = ''
    for c in template:
        if c == '{':
            if expression:
                parts.append(StringPart(expression))
                expression = ''
        elif c == '}':
            variable = expression.split('.', 1)[0].split('[', 1)[0]
            if not variable in names:
                raise TemplateError('Unknown variable "%s" in template "%s".' % (expression, template))
            parts.append(VariablePart(expression))
            expression = ''
        else:
            expression += c
    if expression:
        parts.append(StringPart(expression))
    return Template(template, parts)

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base import NOTSET
from decorated.base.context import Context, ctx
from decorated.base.function import Function, partial, WrapperFunction, \
    ContextFunction
from decorated.decorators import events
from decorated.decorators.conditional import Conditional
from decorated.decorators.events import Event
from decorated.decorators.files import TempFile, TempDir, WritingFile
from decorated.decorators.instantiate import Instantiate
from decorated.decorators.once import Once, OnceSession
from decorated.decorators.profile import Profile
from decorated.decorators.remove_extra_args import RemoveExtraArgs
from decorated.decorators.retries import Retries
from decorated.decorators.synchronized import Synchronized
from decorated.decorators.timeit import TimeIt
from decorated.decorators.timeout import Timeout

NOTSET = NOTSET

Contxt = Context
ctx = ctx

Function = Function
WrapperFunction = WrapperFunction
ContextFunction = ContextFunction
partial = partial

conditional = Conditional
event = events.event
Event = Event
instantiate = Instantiate
once = Once
OnceSession = OnceSession
profile = Profile
remove_extra_args = RemoveExtraArgs
retries = Retries
synchronized = Synchronized
tempfile = TempFile
tempdir = TempDir
timeit = TimeIt
timeout = Timeout
writing_file = WritingFile



# -*- coding: UTF-8 -*-
from decorated.base.proxy import Proxy, NoTargetError
from decorated.util import templates
import doctest
import functools
import inspect

WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__doc__', '__code__', 'func_code')

class ArgError(Exception):
    def __init__(self, param, message):
        super(ArgError, self).__init__(message)
        self.param = param

class Function(Proxy):
    def __init__(self, *args, **kw):
        self.params = None
        self.required_params = None
        self.optional_params = None
        self._func = None
        self._decorate_or_call = self._decorate
        if self._is_init_args(*args, **kw):
            self._init(*args, **kw)
        else:
            self._init()
            self._decorate(args[0])
            
    def __call__(self, *args, **kw):
        return self._decorate_or_call(*args, **kw)  
    
    def __get__(self, obj, cls):
        if obj is None:
            method = Function(self)
        else:
            method = partial(Function, call_args=(obj,))(self)
        method.im_class = cls
        method.im_func = method.__func__ = self
        method.im_self = method.__self__ = obj
        return method
    
    def __str__(self):
        return '<Function %s.%s>' % (self._func.__module__, self.__name__)
    
    @property
    def func(self):
        return self._func.func if isinstance(self._func, Function) else self._func
    
    @property
    def target(self):
        if self._func is None:
            raise NoTargetError()
        return self._func
    
    def _call(self, *args, **kw):
        return self._func(*args, **kw)
    
    def _compile_template(self, template):
        return templates.compile(template, self.params)
    
    def _decorate(self, func):
        self._func = func
        functools.update_wrapper(self, func, WRAPPER_ASSIGNMENTS, updated=())
        if isinstance(func, Function):
            self.params = func.params
            self.required_params = func.required_params
            self.optional_params = func.optional_params
        else:
            self._parse_params(func)
        self._decorate_or_call = self._call
        return self
    
    def _evaluate_expression(self, expression, *args, **kw):
        d = self._resolve_args(*args, **kw)
        return eval(expression, d)
        
    def _init(self, *args, **kw):
        pass
    
    def _is_init_args(self, *args, **kw):
        return len(args) != 1 or not callable(args[0]) or len(kw) != 0
        
    def _parse_params(self, func):
        self.params, _, _, defaults = inspect.getargspec(func)
        if defaults:
            self.required_params = self.params[:-len(defaults)]
            self.optional_params = []
            for i in range(len(defaults) - 1, -1, -1):
                self.optional_params.append((self.params[-1 - i], defaults[-1 - i]))
        else:
            self.required_params = self.params
            self.optional_params = ()
        if _is_bound_method(func):
            self.params = self.params[1:]
            self.required_params = self.required_params[1:]
        self.params = tuple(self.params)
        self.required_params = tuple(self.required_params)
        self.optional_params = tuple(self.optional_params)
        
    def _resolve_args(self, *args, **kw):
        d = {name: default for name, default in self.optional_params}
        for param, arg in zip(self.params, args):
            d[param] = arg
        d.update(kw)
        for name in self.params:
            if name not in d:
                msg = 'Missing argument "%s" for %s.' % (name, str(self))
                raise ArgError(name, msg)
        d = {k: v for k, v in d.items() if k in self.params}
        return d
    
class WrapperFunction(Function):
    def _call(self, *args, **kw):
        self._before(*args, **kw)
        try:
            result = super(WrapperFunction, self)._call(*args, **kw)
        except Exception as e:
            self._error(e, *args, **kw)
            raise
        else:
            self._after(result, *args, **kw)
        return result
    
    def _after(self, ret, *args, **kw):
        pass
    
    def _before(self, *args, **kw):
        pass
    
    def _error(self, error, *args, **kw):
        pass
    
class ContextFunction(WrapperFunction):
    def __enter__(self):
        self._before()
        return self
    
    def __exit__(self, error_type, error_value, traceback):
        if error_value is None:
            self._after(None)
        else:
            self._error(error_value)
    
def partial(func, init_args=(), init_kw=None, call_args=(), call_kw=None):
    if init_kw is None:
        init_kw = {}
    if call_kw is None:
        call_kw = {}
    class _PartialFunction(func):
        def __getattr__(self, name):
            attr = getattr(self._func, name)
            if callable(attr):
                method = attr
                def _wrapper(*args, **kw):
                    args = tuple(call_args) + args
                    merged_kw = dict(call_kw)
                    merged_kw.update(kw)
                    return method(*args, **merged_kw)
                attr = _wrapper
            return attr
        
        def __str__(self):
            return '<PartialFunction %s.%s>' % (self._func.__module__, self.__name__)
    
        def _init(self, *args, **kw):
            args = tuple(init_args) + args
            kw.update(init_kw)
            super(_PartialFunction, self)._init(*args, **kw)
            
        def _call(self, *args, **kw):
            args = tuple(call_args) + args
            merged_kw = dict(call_kw)
            merged_kw.update(kw)
            return super(_PartialFunction, self)._call(*args, **merged_kw)
            
        def _parse_params(self, func):
            super(_PartialFunction, self)._parse_params(func)
            self.params = self.params[len(call_args):]
            self.required_params = self.params[len(call_args):]
            self.params = tuple([p for p in self.params if p not in call_kw])
            self.required_params = tuple([p for p in self.required_params if p not in call_kw])
            self.optional_params = tuple([(k, v) for (k, v) in self.optional_params if k not in call_kw])
    return _PartialFunction
        
def _is_bound_method(func):
    '''
    >>> def foo():
    ...     pass
    >>> _is_bound_method(foo)
    False
    
    >>> class Foo(object):
    ...     def bar(self):
    ...         pass
    >>> _is_bound_method(Foo.bar)
    False
    >>> _is_bound_method(Foo().bar)
    True
    '''
    return hasattr(func, '__self__') and func.__self__ is not None

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base.function import ContextFunction
import os
import shutil

class TempObject(ContextFunction):
    def _after(self, ret, *args, **kw):
        if self._delete_on_success:
            self._do_delete(*args, **kw)
        
    def _calc_path(self, *args, **kw):
        if self._func is None:
            return self._path
        else:
            d = self._resolve_args(*args, **kw)
            return self._path.eval(d)
                
    def _decorate(self, func):
        super(TempObject, self)._decorate(func)
        self._path = self._compile_template(self._path)
        return self
    
    def _do_delete(self, *args, **kw):
        path = self._calc_path(*args, **kw)
        if not os.path.exists(path):
            return
        self._delete(path)
    
    def _error(self, err, *args, **kw):
        if self._delete_on_error:
            self._do_delete(*args, **kw)
        
    def _init(self, path, delete_on_success=True, delete_on_error=True):
        super(TempObject, self)._init()
        self._path = path
        self._delete_on_success = delete_on_success
        self._delete_on_error = delete_on_error
        
class TempFile(TempObject):
    def _delete(self, path):
        os.remove(path)
            
class TempDir(TempObject):
    def _delete(self, path):
        shutil.rmtree(path)

class WritingFile(ContextFunction):
    def discard(self):
        self._discarded = True
        
    def _init(self, path):
        super(WritingFile, self)._init()
        self.path = path
        self.writing_path = self.path + '.writing'
        self._discarded = False
        
    def _after(self, ret, *args, **kw):
        if self._discarded:
            os.remove(self.writing_path)
        else:
            shutil.move(self.writing_path, self.path)
    
    def _error(self, err, *args, **kw):
        os.remove(self.writing_path)
    


# -*- coding: utf-8 -*-
import doctest
import importlib
import pkgutil
import sys

def load_tree(root):
    mod_or_pack = importlib.import_module(root)
    if hasattr(mod_or_pack, '__path__'):
        for _, mod, _ in pkgutil.walk_packages(path=mod_or_pack.__path__, prefix=mod_or_pack.__name__ + '.'):
            if mod not in sys.modules:
                importlib.import_module(mod)
                
def module_exists(modname):
    '''
    >>> module_exists('decorated.util.modutil')
    True
    >>> module_exists('fakepackage.fakemod')
    False
    '''
    try:
        importlib.import_module(modname)
        return True
    except ImportError:
        return False
    
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base.context import Context, ctx, ContextError
from decorated.base.function import Function

class Once(Function):
    def _call(self, *args, **kw):
        key = self._evaluate_expression(self._key, *args, **kw)
        key = (self._func, key)
        try:
            results = ctx._once
        except ContextError:
            results = _DEFAULT_SESSION._once
        except AttributeError:
            results = _DEFAULT_SESSION._once
        if key in results:
            result = results[key]
        else:
            result = super(Once, self)._call(*args, **kw)
            results[key] = result
        return result
    
    def _init(self, key='None'):
        super(Once, self)._init()
        self._key = key
        
class OnceSession(Context):
    def __init__(self):
        super(OnceSession, self).__init__()
        self._once = {}
        
_DEFAULT_SESSION = OnceSession()



# -*- coding: utf-8 -*-
import gc

class DisableGc(object):
    def __enter__(self):
        self._enabled = gc.isenabled()
        gc.disable()
        
    def __exit__(self, *args, **kw):
        if self._enabled:
            gc.enable()
            


# -*- coding: utf-8 -*-

class Proxy(object):
    def __init__(self, target=None):
        super(Proxy, self).__init__()
        self._target = target
        
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            try:
                return getattr(self.target, name)
            except NoTargetError:
                raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
            except AttributeError:
                raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
            
    @property
    def target(self):
        if self._target is None:
            raise NoTargetError()
        return self._target

class NoTargetError(Exception):
    pass



# -*- coding: utf-8 -*-
from decorated.base.function import Function

class Instantiate(Function):
    def _decorate(self, cls):
        self.class_ = cls
        instance = cls()
        method = getattr(instance, self._method)
        def _func(*args, **kw):
            return method(*args, **kw)
        _func.__name__ = cls.__name__
        _func.__module__ = cls.__module__
        return super(Instantiate, self)._decorate(_func)
    
    def _init(self, method='__call__'):
        super(Instantiate, self)._init()
        self._method = method
        


# -*- coding: utf-8 -*-
from decorated.base import NOTSET
from decorated.base.function import WrapperFunction
from decorated.decorators.once import Once
from decorated.decorators.remove_extra_args import RemoveExtraArgs
from decorated.util import modutil
from six import with_metaclass
import doctest

class EventMetaType(type):
    def __init__(self, name, bases, attrs):
        super(EventMetaType, self).__init__(name, bases, attrs)
        self._sources = []
        self._after_listeners = []
        self._before_listeners = []
        
    @property
    def after(self):
        class _EventListener(AfterEventListener):
            _event_class = self
        return _EventListener

    @property
    def before(self):
        class _EventListener(BeforeEventListener):
            _event_class = self
        return _EventListener

    @property
    def name(self):
        return _get_full_name(self)
    
class Event(with_metaclass(EventMetaType, WrapperFunction)):
    fields = ()
    after_fields = ()
    
    @classmethod
    def fire(cls, data=None):
        data = data or {}
        cls._execute_before_listeners(data)
        cls._execute_after_listeners(data)
        
    @classmethod
    def _execute_after_listeners(cls, data):
        for listener in cls._after_listeners:
            listener._call(**data)
            
    @classmethod
    def _execute_before_listeners(cls, data):
        for listener in cls._before_listeners:
            listener._call(**data)
            
    def _after(self, ret, *args, **kw):
        data = self._get_field_values(ret, *args, **kw)
        self._execute_after_listeners(data)
        
    def _before(self, *args, **kw):
        data = self._get_field_values(NOTSET, *args, **kw)
        self._execute_before_listeners(data)
    
    def _decorate(self, func):
        super(Event, self)._decorate(func)
        self._validate()
        type(self)._sources.append(self)
        return self
    
    def _get_field_values(self, ret, *args, **kw):
        args = self._resolve_args(*args, **kw)
        values = {k: args[k] for k in self._basic_fields}
        for key, expression in self._complex_fields.items():
            values[key] = eval(expression, args)
        if ret != NOTSET:
            args['ret'] = ret
            for key, expression in self._after_complex_fields.items():
                values[key] = eval(expression, args)
        return values
    
    def _init(self, **complex_fields):
        super(Event, self)._init()
        self._complex_fields = {k: v for k, v in complex_fields.items() if k in self.fields}
        self._after_complex_fields = {k: v for k, v in complex_fields.items() if k in self.after_fields}
        self._basic_fields = [f for f in self.fields if f not in complex_fields]
    
    def _validate(self):
        for field in self._basic_fields:
            if field not in self.params:
                raise EventError('Missing field "%s" in "%s".' % (field, type(self).name))
        
class EventListener(RemoveExtraArgs):
    def _decorate(self, func):
        super(EventListener, self)._decorate(func)
        self._validate()
        self._register()
        return self
    
    def _register(self):
        raise NotImplemented()
    
    def _validate(self):
        available_fields = self._get_available_fields()
        for param in self.params:
            if param not in available_fields:
                raise EventError('Event "%s" does not have field "%s".' % (self._event_class.name, param))
            
class AfterEventListener(EventListener):
    def _get_available_fields(self):
        return self._event_class.fields + self._event_class.after_fields
    
    def _register(self):
        self._event_class._after_listeners.append(self)
    
class BeforeEventListener(EventListener):
    def _get_available_fields(self):
        return self._event_class.fields
    
    def _register(self):
        self._event_class._before_listeners.append(self)
    
class EventError(Exception):
    pass

def event(fields, after_fields=()):
    fields_ = fields
    after_fields_ = after_fields
    class _Event(Event):
        fields = fields_
        after_fields = after_fields_
    return _Event

@Once
def init(*packages):
    for p in packages:
        modutil.load_tree(p)

def _get_full_name(func):
    '''
    >>> from decorated.util import modutil
    >>> _get_full_name(modutil.load_tree)
    'decorated.util.modutil.load_tree'
    '''
    return '%s.%s' % (func.__module__, func.__name__)

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
import doctest

class Dict(dict):
    '''
    >>> d = Dict(a=1)
    
    >>> 'a' in d
    True
    >>> d['a']
    1
    >>> d.a
    1
    
    >>> d.a = 2
    >>> d['a']
    2
    >>> d.a
    2
    
    >>> d['b'] = 2
    >>> 'b' in d
    True
    >>> d['b']
    2
    >>> d.b
    2
    
    >>> d.c = 3
    >>> 'c' in d
    True
    >>> d['c']
    3
    >>> d.c
    3
    
    >>> del d.c
    >>> 'c' in d
    False
    >>> d['c']
    Traceback (most recent call last):
    ...
    KeyError: 'c'
    >>> d.c
    Traceback (most recent call last):
    ...
    AttributeError: 'c'
    
    >>> del d.d
    Traceback (most recent call last):
    ...
    AttributeError: 'd'
    '''
    def __getattr__(self, name):
        try:
            return super(Dict, self).__getitem__(name)
        except KeyError as e:
            raise AttributeError(str(e))
    
    def __setattr__(self, name, value): 
        super(Dict, self).__setitem__(name, value)
    
    def __delattr__(self, name):
        try:
            super(Dict, self).__delitem__(name)
        except KeyError as e:
            raise AttributeError(str(e))
        
class DefaultDict(Dict):
    '''
    >>> d = DefaultDict(lambda k: 2 * k)
    >>> d['a']
    'aa'
    >>> 'a' in d
    True
    >>> d.b
    'bb'
    >>> 'b' in d
    True
    
    >>> d = DefaultDict()
    >>> d.a is None
    True
    
    >>> d = DefaultDict(a=1, b=2)
    >>> d.a
    1
    
    >>> d = DefaultDict()
    >>> len(d)
    0
    >>> len(d.items())
    0
    >>> len(d.keys())
    0
    >>> len(d.values())
    0
    >>> '_default' in d
    False
    '''
    def __init__(self, default=None, **kw):
        super(DefaultDict, self).__init__(**kw)
        dict.__setattr__(self, '_default', default or (lambda key: None))
        
    def __getattr__(self, key):
        return self.__getitem__(key)
    
    def __getitem__(self, key):
        if key not in self:
            self[key] = self._default(key)
        return super(DefaultDict, self).__getitem__(key)
        
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
import time

from decorated.base.function import Function
from decorated.util import templates, modutil

class CacheDecorator(Function):
    @property
    def invalidate(self):
        return self._cache.uncache(self._key, **self._vars)
    
    def _call(self, *args, **kw):
        d = dict(self._vars)
        d.update(self._resolve_args(*args, **kw))
        key = self._key_template.eval(d)
        return self._process(key, *args, **kw)
        
    def _decorate(self, func):
        super(CacheDecorator, self)._decorate(func)
        var_names = self.params + tuple(self._vars.keys())
        self._key_template = templates.compile(self._key, var_names)
        return self

    def _init(self, cache, key, vars=None, options=None):
        super(CacheDecorator, self)._init()
        self._cache = cache
        self._key = key
        self._vars = vars or {}
        self._options = options
        
    def _process(self, key, *args, **kw):
        result = self._cache._get(key)
        if result is None:
            result = Function._call(self, *args, **kw)
            self._cache._set(key, result)
        return result
    
class Cache(object):
    def cache(self, key, vars=None, **options):
        class _Decorator(CacheDecorator):
            def _process(self, key, *args, **kw):
                result = self._cache._get(key, options)
                if result is None:
                    result = Function._call(self, *args, **kw)
                    self._cache._set(key, result, options)
                return result
        return _Decorator(self, key, vars=vars, options=options)
    
    def uncache(self, key, vars=None, **options):
        class _Decorator(CacheDecorator):
            def _process(self, key, *args, **kw):
                result = Function._call(self, *args, **kw)
                self._cache._delete(key, options)
                return result
        return _Decorator(self, key, vars=vars, options=options)
    
    def _delete(self, key):
        raise NotImplementedError()
    
    def _get(self, key):
        raise NotImplementedError()
    
    def _set(self, key, value):
        raise NotImplementedError()

class SimpleCache(Cache):
    def __init__(self):
        self._data = {}
        
    def _delete(self, key, options):
        del self._data[key]
        
    def _get(self, key, options):
        value, expires = self._data.get(key, (None, None))
        if expires is None or expires < time.time():
            return None
        else:
            return value
        
    def _set(self, key, value, options):
        expires = time.time() + options.get('ttl', 3600)
        self._data[key] = (value, expires)
        
if modutil.module_exists('pylru'):
    from pylru import lrucache
    
    class LruCache(Cache):
        def __init__(self, size=1000):
            self._cache = lrucache(size)
            
        def _delete(self, key, options):
            del self._cache[key]
        
        def _get(self, key, options):
            try:
                return self._cache[key]
            except KeyError:
                return None
        
        def _set(self, key, value, options):
            self._cache[key] = value
            


# -*- coding: utf-8 -*-
from decorated.base.function import Function
from threading import RLock
import os

class Synchronized(Function):
    def _init(self, lock):
        self._lock = lock
        
    def _call(self, *args, **kw):
        with self._lock:
            return super(Synchronized, self)._call(*args, **kw)

MemoryLock = RLock

class FileLock(object):
    def __init__(self, path):
        self._path = path
        
    def __enter__(self):
        import fcntl
        _create_file_if_not_exist(self._path)
        self._fd = os.open(self._path, os.O_RDWR)
        fcntl.flock(self._fd, fcntl.LOCK_EX)
        return self
    
    def __exit__(self, t, v, tb):
        import fcntl
        fcntl.flock(self._fd, fcntl.LOCK_UN)
    
def _create_file_if_not_exist(path):
    try:
        f = open(path, 'w')
        f.close()
    except OSError:
        pass
    


# -*- coding: utf-8 -*-
from distutils.core import setup
import setuptools

setup(
    name='decorated',
    version='1.6.11',
    author='Mengchen LEE',
    author_email='CooledCoffee@gmail.com',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    ],
    description='Decorator framework and common decorators for python.',
    extras_require={
        'test': ['pylru'],
    },
    install_requires=[
        'fixtures2>=0.1.2',
        'six',
    ],
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    url='https://package-insights.appspot.com/packages/decorated',
)



# -*- coding: utf-8 -*-
from decorated.base.dict import Dict
from decorated.base.proxy import Proxy
from decorated.base.thread_local import ThreadLocal
from six import with_metaclass
import doctest
import logging

log = logging.getLogger(__name__)

class ContextMeta(type):
    def __new__(self, name, bases, attrs):
        cls = type.__new__(self, name, bases, attrs)
        sub_class_enter = getattr(cls, '__enter__', None)
        def _enter(instance):
            result = instance
            if sub_class_enter is not None:
                result = sub_class_enter(instance)
            cls._current.set(instance)
            return result
        cls.__enter__ = _enter
        sub_class_exit = getattr(cls, '__exit__', None)
        def _exit(instance, *args, **kw):
            cls._current.set(None)
            if sub_class_exit is not None:
                sub_class_exit(instance, *args, **kw)
        cls.__exit__ = _exit
        cls._current = ThreadLocal()
        return cls
        
    def current(self):
        return self._current.get()
    
class Context(with_metaclass(ContextMeta, Dict)):
    def __init__(self, **kw):
        super(Context, self).__init__(**kw)
        self._parent = None
        
    def __contains__(self, name):
        raise NotImplementedError()
        
    def __enter__(self):
        self._parent = Context.current()
        self._defers = []
        return self
    
    def __exit__(self, error_type, error_value, traceback):
        for defer in self._defers:
            try:
                defer()
            except Exception:
                log.warn('Failed to execute defer "%s".' % defer)
                pass
        Context._current.set(self._parent)
        
    def __getattr__(self, name):
        try:
            return super(Context, self).__getattr__(name)
        except AttributeError:
            if self._parent is None:
                raise
            else:
                try:
                    return getattr(self._parent, name)
                except AttributeError:
                    raise
            
    def __getitem__(self, name):
        raise NotImplementedError()
    
    def defer(self, action):
        self._defers.append(action)
        
    def dict(self):
        data = self._parent.dict() if self._parent else Dict()
        data.update({k: v for k, v in self.items() if not k.startswith('_')})
        return data
        
    def get(self, name, default=None):
        raise NotImplementedError()
            
class ContextProxy(Proxy):
    def __init__(self):
        super(ContextProxy, self).__init__()
        self._inited = True
        
    def __setattr__(self, name, value):
        if self.__dict__.get('_inited'):
            return setattr(self.target, name, value)
        else:
            self.__dict__[name] = value
    
    @property
    def target(self):
        ctx = Context._current.get()
        if ctx:
            return ctx
        else:
            raise ContextError('Context should be set first.')
    
    def get(self):
        return self.target
        
class ContextError(Exception):
    pass

ctx = ContextProxy()

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.base.function import Function

class RemoveExtraArgs(Function):
    def _call(self, *args, **kw):
        kw = self._resolve_args(*args, **kw)
        return super(RemoveExtraArgs, self)._call(**kw)
    


# -*- coding: utf-8 -*-
from decorated.base.function import Function
from decorated.decorators.remove_extra_args import RemoveExtraArgs
import six

class Conditional(Function):
    def _call(self, *args, **kw):
        if self._test(*args, **kw):
            return super(Conditional, self)._call(*args, **kw)
        
    def _init(self, condition=None):
        if hasattr(self, '_test'):
            _test = RemoveExtraArgs(self._test)
        elif isinstance(condition, six.string_types):
            def _test(*args, **kw):
                d = self._resolve_condition_args(*args, **kw)
                return eval(condition, d)
        elif callable(condition):
            condition = RemoveExtraArgs(condition)
            def _test(*args, **kw):
                d = self._resolve_condition_args(*args, **kw)
                return condition(**d)
        else:
            raise Exception('Condition can only be string or callable.')
        self._test = _test
        
    def _resolve_condition_args(self, *args, **kw):
        d = self._resolve_args(*args, **kw)
        if 'self' in d:
            d['self_'] = d.pop('self')
        return d
    


# -*- coding: utf-8 -*-
from fixtures._fixtures.tempdir import TempDir
from fixtures2 import TestCase

class TestCase(TestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self.tempdir = self.useFixture(TempDir())
        


# -*- coding: utf-8 -*-
from decorated.base.function import Function
from fixtures2 import PatchesFixture

class DecoratedFixture(PatchesFixture):
    def disable(self, cls):
        def _call(self, *args, **kw):
            if self._func is None:
                return self._decorate(args[0])
            else:
                return self._func(*args, **kw)
        self.patch_object(cls, '_old__call__', cls.__call__)
        self.patch_object(cls, '__call__', Function._call)
    
    def enable(self, cls):
        cls.__call__ = cls._old__call__